package middleware

import (
	"errors"
	"fmt"
	"io"
	"mime/multipart"
	"net/http"
	"strings"

	"github.com/gin-gonic/gin"
	"hello-api/internal/app/parsers/devicerequestheader"
	"hello-api/internal/logger"
)

const (
	VersionName        = "version"
	lastUpdateTimeName = "LastUpdateTime"
	lastQueryTimeName  = "LastQueryTime"
)

// DeviceRequestHeaderMiddleware -- Reads the body of the context's request, which is assumed to be bytes from an encoded
// msgpack object.
func DeviceRequestHeaderMiddleware(context *gin.Context) {
	contextHeader, err := getMsgPackReader(context.Request)
	if err != nil {
		_ = context.AbortWithError(http.StatusInternalServerError, fmt.Errorf("failed to get the device request header: %w", err))
		return
	} else if contextHeader == nil {
		_ = context.AbortWithError(http.StatusInternalServerError, errors.New("device request header was nil"))
		return
	}
	buf := strings.Builder{}
	_, err = io.Copy(&buf, contextHeader)
	if err != nil {
		_ = context.AbortWithError(http.StatusInternalServerError, fmt.Errorf("failed to read the device request header: %w", err))
		return
	}

	parser := devicerequestheader.NewRequestHeaderDecoder()
	version, lastUpdateTime, lastLiveQueryTime, err := parser.ParseMsgPack(strings.NewReader(buf.String()))
	if err != nil {
		_ = context.AbortWithError(http.StatusInternalServerError, fmt.Errorf("failed to parse the device request header msgpack: %v", err))
		return
	}
	logger.BaseLogger().Debug(fmt.Sprintf("Device request header info(Parsed): version %d, lastUpdateTime %v, lastLiveQueryTime %v",
		version, lastUpdateTime, lastLiveQueryTime))

	context.Set(VersionName, version)
	context.Set(lastUpdateTimeName, lastUpdateTime)
	context.Set(lastQueryTimeName, lastLiveQueryTime)
	context.Next()
}

func getMsgPackReader(r *http.Request) (reader io.Reader, err error) {
	if !strings.Contains(r.Header.Get("Content-Type"), "multipart/form-data") {
		reader = r.Body
		return
	}

	// TODO: is multipart even used with hello?  I could not find any code in CBD

	var multipartReader *multipart.Reader
	multipartReader, err = r.MultipartReader()
	if err != nil {
		return
	}

	// parse form until message field
	part, err := multipartReader.NextPart()
	for err == nil && part.FormName() != "message" {
		part, err = multipartReader.NextPart()
	}
	if err != nil {
		return
	}
	reader = part
	return
}
