package middleware

import (
	"fmt"
	"net/http"

	"carbonblack.com/protocols/psc/ems/server/deviceauth"
	"go.uber.org/zap"

	"hello-api/internal/logger"

	"hello-api/internal/hinting"
	"hello-api/internal/utils"

	"github.com/gin-gonic/gin"
	"github.com/opentracing/opentracing-go"
	"github.com/opentracing/opentracing-go/ext"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints"
)

const HintsName = "hints"

var EmptyHintResult = &hints.HintResult{}

func HintingMiddleware(hintsClient hinting.Hinter) gin.HandlerFunc {
	return func(context *gin.Context) {
		func() {
			var err error
			sp := opentracing.SpanFromContext(context.Request.Context())
			var childSp opentracing.Span = nil
			if sp != nil {
				childSp = opentracing.StartSpan("hint-retrieval", opentracing.ChildOf(sp.Context()))
				defer childSp.Finish()
				defer func() {
					childSp.SetTag(string(ext.Error), err != nil)
				}()
			}
			deviceContext, err := utils.ContextGet[*deviceauth.Context](context, DeviceContextHeader)
			if err != nil {
				_ = context.AbortWithError(http.StatusInternalServerError, fmt.Errorf("failed to get device context header: %w", err))
				return
			}

			// Note: These values provided in 'device_request_header_middleware.go'
			version := context.GetInt(VersionName)
			lastUpdateTime := utils.ContextGetUnixMilliFromPtr(context, lastUpdateTimeName)
			lastQueryTime := utils.ContextGetUnixMilliFromPtr(context, lastQueryTimeName)

			logger.BaseLogger().Debug(fmt.Sprintf("Request sent to GetHints: OrgKey %s, DeviceGroupId %d, DeviceId %d, DeviceType %s, DeploymentType %s",
				deviceContext.Organization.OrgKey, deviceContext.DeviceGroup.Id, deviceContext.Device.Id,
				deviceContext.Device.DeviceType.String(), deviceContext.Device.DeploymentType.String()))

			result, err := hintsClient.GetHints(&hints.JsonHintRequest{
				OrgKey:              deviceContext.Organization.OrgKey,
				DeviceGroupId:       deviceContext.DeviceGroup.Id,
				DeviceId:            deviceContext.Device.Id,
				DeviceType:          deviceContext.Device.DeviceType.String(),
				DeploymentType:      deviceContext.Device.DeploymentType.String(),
				Version:             &version,
				LastUpdateTime:      lastUpdateTime,
				LastLiveQueryTime:   lastQueryTime,
				RemoteRulesRequired: false,
				MessageCode:         nil,
			})
			if err != nil {
				result = EmptyHintResult
				logger.RequestLoggerFromContext(context.Request.Context()).Error("hint_middleware_error", zap.NamedError("hint_retrieval_error", err))
			}
			context.Set(HintsName, result)
			context.Set(VersionName, version)
		}()
		context.Next()
	}
}
