// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package hinting

import (
	"bytes"
	"encoding/json"
	"errors"
	"io"
	"net/http"
	"net/http/httptest"
	"testing"
	"time"

	"hello-api/internal/config"

	"gitlab.bit9.local/psc/ems/hinting/pkg/hints"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints/hinttype"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

func TestHintingClient(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "Hinting client tests")
}

const (
	orgkey = "7ZP4H5XR" // env = "local", orgId = 2023
)

var _ = Describe("Hinting client tests", func() {
	hr := hints.NewHintResult(hinttype.NEW_CONFIG_DATA, hinttype.NEW_POLICY, hinttype.LIVE_QUERY)
	lastUpdateTime := time.UnixMilli(1675397106000)    // Friday, February 3, 2023 4:05:06
	lastLiveQueryTime := time.UnixMilli(1672628645000) // Monday, January 2, 2023 3:04:05
	hr.LastUpdateTime = &lastUpdateTime
	hr.LastLiveQueryTime = &lastLiveQueryTime

	server := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		jsonPayload, _ := json.Marshal(hr)
		w.WriteHeader(http.StatusOK)
		//goland:noinspection GoUnhandledErrorResult
		w.Write(jsonPayload)
	}))

	// ----- Begin Tests --------------------------------------------------

	version := 18
	jsonHintRequest := hints.JsonHintRequest{OrgKey: orgkey, DeviceGroupId: 123, DeviceId: 456, DeviceType: "LINUX", DeploymentType: "ENDPOINT", Version: &version}

	It("get new hints client from config", func() {
		cfg := config.GetInstance()
		cfg.HintingUrl = server.URL
		hintsClientCfg := NewHintsClientFromConfig(cfg)

		hintResult, err := hintsClientCfg.GetHints(&jsonHintRequest)
		Expect(err).To(BeNil())
		Expect(hintResult).ToNot(BeNil())
	})

	hintsClient := NewHintsClient(http.DefaultClient.Do, server.URL)

	It("processes input correctly", func() {
		hintResult, err := hintsClient.GetHints(&jsonHintRequest)
		Expect(err).To(BeNil())

		Expect(hintResult).ToNot(BeNil())
		Expect(hintResult.Hints).To(BeEquivalentTo(hr.Hints))
		Expect(hintResult.LastUpdateTime).To(BeEquivalentTo(&lastUpdateTime))
		Expect(hintResult.LastLiveQueryTime).To(BeEquivalentTo(&lastLiveQueryTime))
	})

	// -----

	It("returns an error when no hint result supplied", func() {
		hintResult, err := hintsClient.GetHints(nil)
		Expect(err).ToNot(BeNil())
		Expect(hintResult).ToNot(BeNil())
	})

	// FUTURE: Figure out how to make a test that has an error on marshal of JsonHintRequest (if possible)
	// FUTURE: Figure out how to make a test that has an error on the NewRequest call

	It("returns an error when hint response has an error", func() {
		hrData, _ := hr.MarshalJSON()
		theClient := NewHintsClient(func(req *http.Request) (*http.Response, error) {
			return &http.Response{StatusCode: 200, Body: io.NopCloser(bytes.NewBuffer(hrData))}, errors.New("bad call")
		}, "ignored")

		hintResult, err := theClient.GetHints(&jsonHintRequest)
		Expect(err).ToNot(BeNil())
		Expect(hintResult).ToNot(BeNil())
	})

	It("returns an error when hint response has a 500 code", func() {
		hrData, _ := hr.MarshalJSON()
		theClient := NewHintsClient(func(req *http.Request) (*http.Response, error) {
			return &http.Response{StatusCode: 500, Body: io.NopCloser(bytes.NewBuffer(hrData))}, nil
		}, "ignored")

		hintResult, err := theClient.GetHints(&jsonHintRequest)
		Expect(err).ToNot(BeNil())
		Expect(hintResult).ToNot(BeNil())
	})

	// FUTURE: Figure out how to make a test that has an error when reading the response body

	It("will return an error if the body can't be unmarshaled", func() {
		_, err := NewHintsClient(func(req *http.Request) (*http.Response, error) {
			return &http.Response{StatusCode: 200, Body: io.NopCloser(bytes.NewBufferString("BOGUS"))}, nil
		}, "ignored").GetHints(&jsonHintRequest)
		Expect(err).ToNot(BeNil())
	})
})
