package lct

import (
	"fmt"

	"go.uber.org/zap"
	"hello-api/internal/logger"
	"hello-api/internal/metrics"

	"carbonblack.com/protocols/psc/devicelastcontacttime"
	producer "github.com/zacharyestep/kinesis-producer"
	"google.golang.org/protobuf/proto"
)

type (

	// LctProducer -- Our simple LCT producer interface for publishing messages
	LctProducer interface {
		Publish(lct *devicelastcontacttime.DeviceLastContactTime) error
	}

	// StartStopableLctProducer -- We need extra methods at the service layer to start/stop and checking if the desired stream actually exists...
	StartStopableLctProducer interface {
		Start()
		Stop()
		LctProducer
	}

	FailureHandler func(<-chan error)

	// LctProducer implementation
	lctProducer struct {
		underlyingProducer underlyingProducerIface
		streamName         string
		FailureHandler
	}

	// KPL library did not expose their producer as an interface (BAD FORM!) so we create one in our code to facilitate easy testing
	underlyingProducerIface interface {
		Put(data []byte, partitionKey string) error
		NotifyFailures() <-chan error
		Start()
		Stop()
	}
)

func NewLctProducerFromConfig(producerConfig *producer.Config, metricsClient metrics.Service) (StartStopableLctProducer, error) {
	return NewLctProducer(producerConfig, defaultFailureHandler(metricsClient),
		func(config *producer.Config) (underlyingProducerIface, error) {
			return producer.New(config)
		})
}

func NewLctProducer(producerConfig *producer.Config, failureHandler FailureHandler, producerBuilder func(producerConfig *producer.Config) (underlyingProducerIface, error)) (StartStopableLctProducer, error) {
	kinesisProducer, err := producerBuilder(producerConfig)
	return &lctProducer{underlyingProducer: kinesisProducer, streamName: producerConfig.StreamName, FailureHandler: failureHandler}, err
}

func defaultFailureHandler(metricsClient metrics.Service) FailureHandler {
	return func(records <-chan error) {
		go func() {
			for r := range records {
				logger.BaseLogger().Error("Error while posting to kinesis", zap.Any("failure_record", r))
				if metricsClient != nil {
					metricsClient.Count("kpl.hello-lct-stream.failed", nil, 1)
				}
			}
		}()
	}
}

func (lctPublisher *lctProducer) Start() {
	lctPublisher.FailureHandler(lctPublisher.underlyingProducer.NotifyFailures())
	lctPublisher.underlyingProducer.Start()
}

func (lctPublisher *lctProducer) Stop() {
	lctPublisher.underlyingProducer.Stop()
}

func (lctPublisher *lctProducer) Publish(lct *devicelastcontacttime.DeviceLastContactTime) error {
	bytes, err := proto.Marshal(lct)
	if err != nil {
		return err
	}
	return lctPublisher.underlyingProducer.Put(bytes, fmt.Sprintf("%d", lct.GetDeviceId()))
}
